import time
from asyncio import Future
from collections.abc import Coroutine
from contextlib import asynccontextmanager
from typing import TYPE_CHECKING, Any

import aiohttp
from aiohttp import ClientConnectorError, ClientResponse, ClientResponseError, ClientSession
from aiohttp.client import _RequestContextManager
from opentelemetry import context
from opentelemetry.trace import Span

from aiolocust import User, stats
from aiolocust.datatypes import Request

if TYPE_CHECKING:  # avoid circular import
    from aiolocust.runner import Runner


SPAN_NAME_KEY = context.create_key("name")


class HttpUser(User):
    def __init__(self, runner: Runner | None = None, base_url=None, **kwargs):
        super().__init__(runner)
        self.base_url = base_url
        self.kwargs = kwargs
        self.client: LocustClientSession  # type: ignore[assignment] # always set in cm

    @asynccontextmanager
    async def cm(self):
        async with LocustClientSession(self.runner, self.base_url, **self.kwargs) as self.client:
            yield


class LocustResponse(ClientResponse):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.error: Exception | bool | str | None = None


class LocustRequestContextManager(_RequestContextManager):
    def __init__(self, name, coro: Coroutine[Future[Any], None, ClientResponse]):
        super().__init__(coro)
        # slightly hacky way to get the URL, but passing it explicitly would be a mess
        # and it is only used for connection errors where the exception doesn't contain URL
        self.str_or_url = coro._coro.cr_frame.f_locals["str_or_url"]  # type: ignore
        self._resp: LocustResponse  # type: ignore
        self.name = name

    async def __aenter__(self) -> LocustResponse:
        ctx = context.set_value(SPAN_NAME_KEY, self.name)
        token = context.attach(ctx)
        self.start_time = time.perf_counter()
        try:
            await super().__aenter__()
        except ClientConnectorError as e:
            elapsed = self.ttlb = time.perf_counter() - self.start_time
            if request_info := getattr(e, "request_info", None):
                url = request_info.url
            else:
                url = self.str_or_url
            stats.request(Request(str(self.name or url), elapsed, elapsed, e))
            raise
        except ClientResponseError as e:
            elapsed = self.ttlb = time.perf_counter() - self.start_time
            stats.request(Request(str(self.name or self.str_or_url), elapsed, elapsed, e))
            raise
        except TimeoutError as e:
            elapsed = self.ttlb = time.perf_counter() - self.start_time
            stats.request(Request(str(self.name or self.str_or_url), elapsed, elapsed, e))
            raise
        else:
            self.url = super()._resp.url
            self.ttfb = time.perf_counter() - self.start_time
            await self._resp.read()
            self.ttlb = time.perf_counter() - self.start_time
        finally:
            context.detach(token)

        return self._resp

    async def __aexit__(self, exc_type, exc_val, exc_tb) -> None:
        await super().__aexit__(exc_type, exc_val, exc_tb)
        if self._resp.error is None:  # no explicit value set in with-block
            try:
                self._resp.raise_for_status()
            except (ClientResponseError, ClientConnectorError) as e:
                self._resp.error = e
            if exc_val:  # overwrite if there was an explicit exception (e.g. an assert or crash)
                self._resp.error = exc_val
        stats.request(
            Request(
                str(self.name or self.url),
                self.ttfb,
                self.ttlb,
                self._resp.error,
            )
        )


def request_hook(span: Span, params: aiohttp.TraceRequestStartParams):  # noqa: ARG001
    """
    Request hook for renaming spans based on name parameter (passed via context vars)

    Typical usage:

    AioHttpClientInstrumentor().instrument(request_hook=aiolocust.http.request_hook)
    """
    if custom_name := context.get_value(SPAN_NAME_KEY):
        span.update_name(str(custom_name))


class LocustClientSession(ClientSession):
    def __init__(self, runner: Runner | None = None, base_url=None, **kwargs):
        super().__init__(base_url=base_url, response_class=LocustResponse, **kwargs)
        self.runner: Runner = runner  # pyright: ignore[reportAttributeAccessIssue] # always set outside of unit testing

    # explicitly declare this to get the correct return type
    async def __aenter__(self) -> LocustClientSession:
        return self

    def get(self, url, *, name=None, **kwargs) -> LocustRequestContextManager:
        return LocustRequestContextManager(name, super().get(url, **kwargs))

    def post(self, url, *, name=None, **kwargs) -> LocustRequestContextManager:
        return LocustRequestContextManager(name, super().post(url, **kwargs))

    def options(self, url, *, name=None, **kwargs) -> LocustRequestContextManager:
        return LocustRequestContextManager(name, super().options(url, **kwargs))

    def head(self, url, *, name=None, **kwargs) -> LocustRequestContextManager:
        return LocustRequestContextManager(name, super().head(url, **kwargs))

    def put(self, url, *, name=None, **kwargs) -> LocustRequestContextManager:
        return LocustRequestContextManager(name, super().put(url, **kwargs))

    def patch(self, url, *, name=None, **kwargs) -> LocustRequestContextManager:
        return LocustRequestContextManager(name, super().patch(url, **kwargs))

    def delete(self, url, *, name=None, **kwargs) -> LocustRequestContextManager:
        return LocustRequestContextManager(name, super().delete(url, **kwargs))



import asyncio
import logging
import math
import os
import signal
import sys
import threading
import time
import warnings

from aiohttp import ClientResponseError
from rich.console import Console

from aiolocust import User, otel, stats
from aiolocust.datatypes import Stage

# uvloop is faster than the default pure-python asyncio event loop
# so if it is installed, we're going to be using that one
try:
    import uvloop

    new_event_loop = uvloop.new_event_loop
except ImportError:
    new_event_loop = None

logger = logging.getLogger(__name__)

# Some exceptions will be raised by user code trigger a restart of the run method without propagating it further.
# Gotta do some special logic for Playwright, because it is an optional dependency.
try:
    import playwright.async_api  # pyright: ignore[reportMissingImports]

    EXPECTED_ERRORS = (ClientResponseError, AssertionError, TimeoutError, playwright.async_api.TimeoutError)
except ImportError:
    EXPECTED_ERRORS = (ClientResponseError, AssertionError, TimeoutError)


# We're going to inherit from ClientSession, even though it is considered internal,
# Because we dont want to take the performance hit and typing issues of wrapping every method
warnings.filterwarnings(
    action="ignore",
    message=".*Inheritance .* from ClientSession is discouraged.*",
    category=DeprecationWarning,
    module="aiolocust",
)


if sys._is_gil_enabled():
    # Note: logging has not yet been configured, so it won't get proper formatting
    logger.warning("aiolocust requires a freethreading Python build")


original_sigint_handler = signal.getsignal(signal.SIGINT)


def distribute_evenly(total, num_buckets) -> list[int]:
    # Calculate the base amount for every bucket
    base = total // num_buckets
    # Calculate how many buckets need an extra +1
    remainder = total % num_buckets
    # Create the list: add 1 to the first 'remainder' buckets
    return [base + 1 if i < remainder else base for i in range(num_buckets)]


def desired_user_count(stages: list[Stage], elapsed: float) -> int | None:
    total_time = 0.0
    previous_user_count = 0
    for stage in stages:
        total_time += stage.duration
        if elapsed <= total_time:
            time_left_in_stage = total_time - elapsed
            return math.ceil(
                previous_user_count + (stage.target - previous_user_count) * (1 - time_left_in_stage / stage.duration)
            )
        previous_user_count = stage.target

    return None


class LoopWorker(threading.Thread):
    def __init__(self):
        super().__init__(daemon=True)
        self.loop = asyncio.new_event_loop()

    def run(self):
        asyncio.set_event_loop(self.loop)
        self.loop.run_forever()

    def stop(self):
        self.loop.call_soon_threadsafe(self.loop.stop)


class Runner:
    def __init__(
        self,
        users: list[type[User]],
        user_count: int = 1,
        duration: int | None = None,
        rate: float | None = None,
        config: dict | None = None,
        event_loops: int | None = None,
    ):
        signal.signal(signal.SIGINT, self.signal_handler)
        self.running = False
        self.start_time = 0
        self.sf = stats.StatsFormatter()
        self.console = Console()

        if config and "stages" in config:
            self.stages = [Stage(**item) for item in config["stages"]]
            if user_count > 1 or duration or rate:
                logger.info("Both stages and user_count/duration/rate were specified, stages will take precedence")
        else:
            ramp_up_time = user_count / rate if rate else 0
            self.stages = [
                Stage(ramp_up_time, user_count),
                Stage(duration - ramp_up_time if duration else 99999999, user_count),
            ]
        logger.debug(f"Stages: {self.stages}")

        self.users = users
        if event_loops is None:
            if cpu_count := os.cpu_count():
                # for heavy calculations this may need to be increased,
                # but for I/O bound tasks 1/2 of CPU cores seems to be very efficient
                self.event_loops = max(cpu_count // 2, 1)
            else:
                self.event_loops = 1
        else:
            self.event_loops = event_loops
        self.running_users: set[User] = set()
        self.futures: list[asyncio.Future] = []

    async def stats_printer(self):
        first = True
        while self.running:
            if not first:
                self.console.print(self.sf.get_table())
            first = False
            await asyncio.sleep(2)

    def shutdown(self):
        if not self.running:
            logger.warning("Already shutting down, ignoring shutdown() call")
            return
        self.running = False
        for user in self.running_users:
            user.running = False
        for fut in self.futures:
            _ = fut.result()
        otel.logger_provider.shutdown()

    async def user_loop(self, user_instance: User):
        async with user_instance.cm():
            while user_instance.running:
                try:
                    await user_instance.run()
                except EXPECTED_ERRORS:
                    pass  # these errors should already have been recorded by the User
                except Exception as e:
                    stats.record_error(str(e))
                    logger.exception(e)

    def signal_handler(self, _sig, _frame):
        signal.signal(signal.SIGINT, original_sigint_handler)  # stop immediately on second Ctrl-C
        logger.info("\nStopping...")
        self.shutdown()

    def run_test(self):
        asyncio.run(self.run_test_async())

    def add_user(self, worker: LoopWorker):
        user = self.users[0](self)
        self.running_users.add(user)
        fut = asyncio.run_coroutine_threadsafe(self.user_loop(user), worker.loop)
        self.futures.append(fut)  # type: ignore

    def stop_user(self):
        user = self.running_users.pop()
        user.running = False

    async def run_test_async(self):
        self.running = True

        workers = [LoopWorker() for _ in range(self.event_loops)]
        for w in workers:
            w.start()
        logger.debug(f"Running with {self.event_loops} event loops")
        await asyncio.sleep(0.1)

        loop = asyncio.get_running_loop()
        stats_printer_task = loop.create_task(self.stats_printer())

        self.start_time = time.time()
        self.previous_user_count = 0

        while self.running:
            await asyncio.sleep(0.01)
            elapsed = time.time() - self.start_time
            new_user_count = desired_user_count(self.stages, elapsed)
            if new_user_count is None:
                break
            change = new_user_count - self.previous_user_count
            if change > 0:
                for i in range(change):
                    worker = workers[(i + self.previous_user_count) % self.event_loops]
                    self.add_user(worker)
            elif change < 0:
                for i in range(-change):
                    self.stop_user()
            self.previous_user_count = new_user_count

        if self.running:  # if we exited the loop without a signal, we should still do a proper shutdown
            self.shutdown()

        stats_printer_task.cancel()

        self.console.print(self.sf.get_table(True))
        if stats.error_counter:
            self.console.print(self.sf.get_error_table())

        for w in workers:
            w.stop()

        return



import logging
import os
import sys

from opentelemetry import metrics, trace
from opentelemetry._logs import set_logger_provider
from opentelemetry.instrumentation.logging.handler import LoggingHandler
from opentelemetry.sdk._logs import LoggerProvider
from opentelemetry.sdk._logs.export import BatchLogRecordProcessor, ConsoleLogRecordExporter
from opentelemetry.sdk.metrics import MeterProvider
from opentelemetry.sdk.metrics.export import ConsoleMetricExporter, MetricReader, PeriodicExportingMetricReader
from opentelemetry.sdk.resources import Resource
from opentelemetry.sdk.trace import TracerProvider
from opentelemetry.sdk.trace.export import BatchSpanProcessor, ConsoleSpanExporter, SimpleSpanProcessor
from rich.console import Console
from rich.logging import RichHandler

resource = Resource.create(
    {
        "service.name": "locust",
        "service.version": "0.0.0",  # __version__
    }
)


logger_provider = LoggerProvider(resource=resource)
set_logger_provider(logger_provider)
tracer_provider = TracerProvider(resource=resource)
trace.set_tracer_provider(tracer_provider)
tracer = tracer_provider.get_tracer("aiolocust")


def setup_logging(level: int = logging.INFO):
    otel_handler = LoggingHandler(level=level, logger_provider=logger_provider)

    logs_exporters = {e.strip().lower() for e in os.getenv("OTEL_LOGS_EXPORTER", "otlp").split(",") if e.strip()}
    for exporter in logs_exporters:
        if exporter == "otlp":
            protocol = (
                os.getenv("OTEL_EXPORTER_OTLP_LOGS_PROTOCOL", os.getenv("OTEL_EXPORTER_OTLP_PROTOCOL", "http/protobuf"))
                .lower()
                .strip()
            )
            try:
                if protocol == "grpc":
                    from opentelemetry.exporter.otlp.proto.grpc._log_exporter import OTLPLogExporter  # type: ignore
                elif protocol == "http/protobuf" or protocol == "http":
                    from opentelemetry.exporter.otlp.proto.http._log_exporter import OTLPLogExporter  # type: ignore
                else:
                    print(
                        f"Unknown OpenTelemetry otlp logs exporter protocol '{protocol}'. Use 'grpc' or 'http/protobuf'"
                    )
                    continue
            except ImportError:
                if (
                    level == logging.INFO
                    and os.getenv("OTEL_LOGS_EXPORTER", "")
                    or os.getenv("OTEL_EXPORTER_OTLP_PROTOCOL", "")
                ):
                    print(
                        f"OpenTelemetry otlp exporter for '{protocol}' is not available. Please install the required package: opentelemetry-exporter-otlp-proto-{'grpc' if protocol == 'grpc' else 'http'}",
                    )
                continue

            otlp_exporter = OTLPLogExporter()
            logger_provider.add_log_record_processor(BatchLogRecordProcessor(otlp_exporter))
        elif exporter == "console":
            logger_provider.add_log_record_processor(
                BatchLogRecordProcessor(ConsoleLogRecordExporter(), schedule_delay_millis=100)
            )

        elif exporter == "none":
            continue

        else:
            print(f"Unknown logs exporter '{exporter}'. Ignored")

    # Use RichHandler only when stderr is a TTY (interactive) to avoid double-formatting.
    if sys.stderr.isatty():
        logging.basicConfig(
            handlers=[otel_handler, RichHandler(level, console=Console(stderr=True))],
            datefmt="[%X]",
            level=level,
            format="%(message)s",
        )
    else:
        stream_handler = logging.StreamHandler(sys.stderr)
        stream_handler.setFormatter(logging.Formatter("[%(asctime)s] %(levelname)s/%(name)s: %(message)s"))
        logging.basicConfig(
            handlers=[otel_handler, stream_handler],
            datefmt="[%X]",
            level=level,
            format="%(message)s",
        )


def setup_trace_exporters():
    logger = logging.getLogger(__name__)
    traces_exporters = {e.strip().lower() for e in os.getenv("OTEL_TRACES_EXPORTER", "otlp").split(",") if e.strip()}
    for exporter in traces_exporters:
        if exporter == "otlp":
            protocol = (
                os.getenv(
                    "OTEL_EXPORTER_OTLP_TRACES_PROTOCOL", os.getenv("OTEL_EXPORTER_OTLP_PROTOCOL", "http/protobuf")
                )
                .lower()
                .strip()
            )
            try:
                if protocol == "grpc":
                    from opentelemetry.exporter.otlp.proto.grpc.trace_exporter import OTLPSpanExporter  # type: ignore
                elif protocol == "http/protobuf" or protocol == "http":
                    from opentelemetry.exporter.otlp.proto.http.trace_exporter import OTLPSpanExporter  # type: ignore
                else:
                    print(
                        f"Unknown OpenTelemetry otlp traces exporter protocol '{protocol}'. Use 'grpc' or 'http/protobuf'"
                    )
                    continue
            except ImportError:
                level = logging.INFO if os.getenv("OTEL_TRACES_EXPORTER", "") else logging.DEBUG
                logger.log(
                    level,
                    f"OpenTelemetry otlp exporter for '{protocol}' is not available. Please install the required package: opentelemetry-exporter-otlp-proto-{'grpc' if protocol == 'grpc' else 'http'}",
                )
                continue

            tracer_provider.add_span_processor(BatchSpanProcessor(OTLPSpanExporter()))

        elif exporter == "console":
            tracer_provider.add_span_processor(SimpleSpanProcessor(ConsoleSpanExporter()))

        elif exporter == "none":
            continue

        else:
            print(f"Unknown traces exporter '{exporter}'. Ignored")


def setup_meter_provider(metric_readers: list[MetricReader]):
    logger = logging.getLogger(__name__)
    metrics_exporters = {e.strip().lower() for e in os.getenv("OTEL_METRICS_EXPORTER", "otlp").split(",") if e.strip()}
    for exporter in metrics_exporters:
        if exporter == "otlp":
            protocol = (
                os.getenv(
                    "OTEL_EXPORTER_OTLP_METRICS_PROTOCOL", os.getenv("OTEL_EXPORTER_OTLP_PROTOCOL", "http/protobuf")
                )
                .lower()
                .strip()
            )
            try:
                if protocol == "grpc":
                    from opentelemetry.exporter.otlp.proto.grpc.metric_exporter import (  # type: ignore
                        OTLPMetricExporter,
                    )
                elif protocol == "http/protobuf" or protocol == "http":
                    from opentelemetry.exporter.otlp.proto.http.metric_exporter import (  # type: ignore
                        OTLPMetricExporter,
                    )
                else:
                    print(
                        f"Unknown OpenTelemetry otlp meter exporter protocol '{protocol}'. Use 'grpc' or 'http/protobuf'"
                    )
                    continue
            except ImportError:
                level = logging.INFO if os.getenv("OTEL_METRICS_EXPORTER", "") else logging.DEBUG
                logger.log(
                    level,
                    f"OpenTelemetry otlp exporter for '{protocol}' is not available. Please install the required package: opentelemetry-exporter-otlp-proto-{'grpc' if protocol == 'grpc' else 'http'}",
                )
                continue

            metric_reader = PeriodicExportingMetricReader(OTLPMetricExporter())
            metric_readers.append(metric_reader)

        elif exporter == "prometheus":
            print("Prometheus metrics exporter is not yet implemented!")

        elif exporter == "console":
            metric_reader = PeriodicExportingMetricReader(ConsoleMetricExporter())
            metric_readers.append(metric_reader)

        elif exporter == "none":
            continue

        else:
            print(f"Unknown metrics exporter '{exporter}'. Ignored")

    metrics.set_meter_provider(MeterProvider(resource=resource, metric_readers=metric_readers))



from abc import ABC, abstractmethod
from contextlib import asynccontextmanager


class User(ABC):
    def __init__(self, runner: Runner | None = None, **kwargs):
        self.runner = runner
        self.running = True

    @abstractmethod
    async def run(self): ...

    @asynccontextmanager
    async def cm(self):
        """Override this method if you need an async context manager around the run method"""
        yield


from aiolocust.http import HttpUser, LocustClientSession  # noqa: F401
from aiolocust.runner import Runner  # noqa: F401



import time
from collections import defaultdict
from threading import Lock

from opentelemetry import metrics
from opentelemetry.sdk.metrics import Histogram
from opentelemetry.sdk.metrics.export import AggregationTemporality, HistogramDataPoint, InMemoryMetricReader
from rich.table import Table

from aiolocust import otel
from aiolocust.datatypes import Request, RequestEntry

MAX_ERROR_KEYS = 200

reader = InMemoryMetricReader(
    preferred_temporality={
        Histogram: AggregationTemporality.DELTA,
    }
)

otel.setup_trace_exporters()
otel.setup_meter_provider([reader])
meter = metrics.get_meter("locust")
ttlb_histogram = meter.create_histogram("http.client.duration")
error_counter = defaultdict(int)
error_counter_lock = Lock()


def record_error(message: str):
    with error_counter_lock:
        if message not in error_counter and len(error_counter) >= MAX_ERROR_KEYS:
            message = "OTHER"
        error_counter[message] += 1


def request(req: Request):
    attributes = {
        "http.url": req.url,
        # the rest of these remain to be implemented
        # http.method=GET,
        # http.host=localhost,
        # net.peer.name=localhost,
        # net.peer.port=8080,
        # http.status_code=200}
    }
    if req.error:
        attributes["error.type"] = req.error.__class__.__name__
        record_error(str(req.error) or req.error.__class__.__name__)
    ttlb_histogram.record(req.ttlb, attributes=attributes)


def make_row(name: str, re: RequestEntry, start, end) -> list[str]:
    return [
        name,
        str(re.count),
        f"{re.errorcount} ({re.error_percentage:2.1f}%)",
        f"{re.avg_ttlb_ms:4.1f}ms",
        f"{re.max_ttlb_ms:4.1f}ms",
        f"{re.rate(start, end):.2f}/s",
    ]


class StatsFormatter:
    def __init__(self):
        self.start_time = time.time()
        self.last_time = self.start_time
        self.aggregate: dict[str, RequestEntry] = defaultdict(RequestEntry)
        # clear reader, in case this is not the first Stats object
        _ = reader.get_metrics_data()
        error_counter.clear()

    def _get_entries(self) -> dict[str, RequestEntry]:
        metrics_data = reader.get_metrics_data()
        entries: dict[str, RequestEntry] = defaultdict(RequestEntry)
        for resource_metric in metrics_data.resource_metrics if metrics_data else []:
            for scope_metric in resource_metric.scope_metrics:
                for metric in scope_metric.metrics:
                    for point in metric.data.data_points:
                        if not point.attributes:
                            raise Exception(f"A data point had no attributes, that should never happen. Point: {point}")
                        if not isinstance(point, HistogramDataPoint):
                            continue  # maybe we should log this, but we don't want to kill performance either
                        if name := point.attributes.get("http.url"):
                            entries[str(name)] += RequestEntry(
                                point.count,
                                point.count if point.attributes.get("error.type") else 0,
                                point.sum,
                                point.max,
                            )
                        else:
                            pass  # print(f"Unknown point {point}")

        return entries

    def _get_rows(self, final_summary) -> list[list[str]]:
        table: list[list[str]] = []
        summary_table: list[list[str]] = []
        now = time.time()

        entries = self._get_entries()
        total = RequestEntry()

        for url, re in entries.items():
            self.aggregate[url] += re
            total += re
            table.append(make_row(url, re, self.last_time, now))
        table.append(make_row("Total", total, self.last_time, now))

        self.last_time = now

        if not final_summary:
            return table

        summary_total = RequestEntry()

        for url, re in self.aggregate.items():
            summary_total += re
            summary_table.append(make_row(url, re, self.start_time, now))
        summary_table.append(make_row("Total", summary_total, self.start_time, now))

        return summary_table

    def get_table(self, final_summary=False):
        table = Table(show_edge=False)
        table.add_column("Name", max_width=30)
        table.add_column("Count", justify="right")
        table.add_column("Failures", justify="right")
        table.add_column("Avg", justify="right")
        table.add_column("Max", justify="right")
        table.add_column("Rate", justify="right")

        for row in self._get_rows(final_summary):
            table.add_row(*row)

        if final_summary:
            table.title = "Summary"

        return table

    def get_error_table(self):
        error_table = Table(show_edge=False)
        error_table.add_column("Count")
        error_table.add_column("Error")

        for key, count in sorted(error_counter.items(), key=lambda item: item[1], reverse=True):
            error_table.add_row(str(count), key)

        return error_table



import importlib.util
import inspect
import json
import logging
import os
import sys
from enum import StrEnum
from pathlib import Path
from typing import Annotated

import typer

from aiolocust import User
from aiolocust.http import HttpUser
from aiolocust.otel import setup_logging
from aiolocust.runner import Runner


class LogLevel(StrEnum):
    debug = "debug"
    info = "info"
    warning = "warning"
    error = "error"


app = typer.Typer(add_completion=False)
logger = logging.getLogger(__name__)


def is_user_class(item) -> bool:
    """
    Check if a variable is a runnable (non-abstract) User class
    """
    return bool(inspect.isclass(item)) and issubclass(item, User) and not inspect.isabstract(item)


def load_config(input_string):
    if os.path.isfile(input_string):
        try:
            with open(input_string) as f:
                return json.load(f)
        except json.JSONDecodeError as e:
            print(f"Config file is not valid JSON: {e}")
            raise
        except Exception as e:
            print(f"Could not load file: {e}")
            raise
    try:
        return json.loads(input_string)
    except json.JSONDecodeError as e:
        print(f"Config is not valid JSON: {e}")
        raise


@app.command()
def main(
    filename: Annotated[str, typer.Argument(help="The test to run")] = "locustfile.py",
    users: Annotated[int, typer.Option("-u", "--users", help="The number of concurrent VUs")] = 1,
    duration: Annotated[int | None, typer.Option("-d", "--duration", help="Stop the test after X seconds")] = None,
    rate: Annotated[
        float | None, typer.Option("-r", "--rate", help="Rate to spawn users at (users per second).")
    ] = None,
    log_level: Annotated[
        LogLevel, typer.Option("--log-level", help="Set the logging level", case_sensitive=False)
    ] = LogLevel.info,
    config: Annotated[
        dict | None,
        typer.Option(
            metavar="JSON",
            parser=load_config,
            help='JSON string or path to JSON file, e.g. \n\n{"stages": [{"duration": 10, "target": 10}, {"duration": 5, "target": 0}]}',
        ),
    ] = None,
    event_loops: Annotated[
        int | None,
        typer.Option(
            "--event-loops", help="Set the number of aio event loops", rich_help_panel="Advanced Configuration"
        ),
    ] = None,
):
    log_level_id = getattr(logging, log_level.value.upper())
    setup_logging(log_level_id)

    file_path = Path(filename).resolve()
    if not file_path.exists():
        typer.echo(f"Error: Could not find the file at {file_path}")
        raise typer.Exit(code=1)

    module_name = file_path.stem

    spec = importlib.util.spec_from_file_location(module_name, file_path)
    if spec is None or spec.loader is None:
        typer.echo(f"Error: Could not load the file at {file_path}")
        raise typer.Exit(code=1)

    module = importlib.util.module_from_spec(spec)

    # Add the module to sys.modules so it behaves like a normal import
    sys.modules[module_name] = module

    # Run any top-level code
    spec.loader.exec_module(module)

    # Return our two-tuple
    user_classes = {name: value for name, value in vars(module).items() if is_user_class(value)}
    if not user_classes and hasattr(module, "run"):

        class SimpleUser(HttpUser):
            async def run(self):
                pass  # This will be overwritten immediately, but needs to be here to satisfy the abstract base class requirement

        SimpleUser.run = module.run
        user_classes = {"SimpleUser": SimpleUser}

    if user_classes:
        r = Runner(
            [user for user in user_classes.values()],
            user_count=users,
            duration=duration,
            rate=rate,
            config=config,
            event_loops=event_loops,
        )
        r.run_test()
    else:
        typer.echo(f"Error: No User classes or run function defined in {filename}")



from contextlib import asynccontextmanager

from opentelemetry import trace
from playwright.async_api import Page, async_playwright  # pyright: ignore[reportMissingImports]

from aiolocust import User, stats
from aiolocust.datatypes import Request
from aiolocust.runner import Runner

# Setup OTel Tracer (this is probably going to need to change)
tracer = trace.get_tracer("playwright-instrumentation")

playwright_instance = None
browser_instance = None


class LocustPage:
    """A wrapper for the Playwright Page object to automatically generate OTel spans."""

    def __init__(self, page: Page):
        self._page = page

    async def goto(self, url: str, **kwargs):
        with tracer.start_as_current_span("playwright.goto") as span:
            span.set_attribute("browser.url", url)
            try:
                result = await self._page.goto(url, **kwargs)
                stats.request(Request(url, 1, 1, None))
            except Exception as e:
                span.record_exception(e)
                stats.request(Request(url, 1, 1, e))
                raise
            return result

    async def click(self, selector: str, **kwargs):
        with tracer.start_as_current_span("playwright.click") as span:
            span.set_attribute("browser.selector", selector)
            try:
                result = await self._page.click(selector, **kwargs)
                stats.request(Request(selector, 1, 1, None))
            except Exception as e:
                span.record_exception(e)
                stats.request(Request(selector, 1, 1, e))
                raise
            return result


class PlaywrightUser(User):
    def __init__(self, runner: Runner | None = None, **kwargs):
        super().__init__(runner)
        self.kwargs = kwargs
        self.page: LocustPage  # type: ignore[assignment] # always set in cm

    @asynccontextmanager
    async def cm(self):
        global playwright_instance, browser_instance
        if playwright_instance is None:
            playwright_instance = await async_playwright().start()
            browser_instance = await playwright_instance.chromium.launch(
                headless=False,  # Probably wanna set this to true later on
                args=[
                    "--disable-blink-features=AutomationControlled",
                    "--safebrowsing-disable-auto-update",
                    "--disable-sync",
                    "--hide-scrollbars",
                    "--disable-notifications",
                    "--disable-logging",
                    "--ignore-certificate-errors",
                    "--no-first-run",
                    "--disable-audio-output",
                    "--disable-canvas-aa",
                    "--lang=en-US",
                    "--disable-features=LanguageDetection",
                ],
                handle_sigint=False,
            )
        assert browser_instance
        context = await browser_instance.new_context()
        raw_page = await context.new_page()
        self.page = LocustPage(raw_page)

        yield

        await raw_page.close()
        await context.close()



from dataclasses import dataclass


@dataclass(slots=True)
class Request:
    url: str
    ttfb: float
    ttlb: float
    error: Exception | bool | str | None


@dataclass(slots=True)
class RequestEntry:
    count: int = 0
    errorcount: int = 0
    sum_ttlb: float = 0.0
    max_ttlb: float = 0.0

    def __iadd__(self, other: RequestEntry):
        if isinstance(other, RequestEntry):
            self.count += other.count
            self.errorcount += other.errorcount
            self.sum_ttlb += other.sum_ttlb
            self.max_ttlb = max(self.max_ttlb, other.max_ttlb)
            return self

    def rate(self, start, end) -> float:
        return self.count / (end - start)

    @property
    def avg_ttlb(self) -> float:
        return self.sum_ttlb / self.count if self.count > 0 else 0.0

    @property
    def avg_ttlb_ms(self) -> float:
        return self.avg_ttlb * 1000

    @property
    def max_ttlb_ms(self) -> float:
        return self.max_ttlb * 1000

    @property
    def error_percentage(self) -> float:
        return self.errorcount / self.count * 100.0 if self.count > 0 else 0.0


@dataclass
class Stage:
    duration: float
    target: int
